(** Church-Style Terms for System F omega *)

open Output
open Pie
open Exp
open Surr
open Eq

exception Error of string

let indent = 2

type clos = Clos.surr * Clos.surr * Exp.exp (** Church-style closures *)


(** Operations on Church-style cloures *)

(* unused
let tmAbs' ((thL, gL), (thR, gR), e) = 
  let (gL', e') = tmAbstract (gL, gR, e)
  in ((thL, gL'), e')
*)

(** Printing Church-style terms *)
		  
(* precedence:
   11: tyApp left
   10: app right
    9: app left
    5: -> left
    4: -> right
    0: forall, lam
*)
	
let toString ((thL, gL0), (thR, gR0), e0) =
  let rec tos (b, n, gL, gR , e) =
    let par () = 
      let s = tos (true, 0, gL, gR, e) in
	if !Output.extensiveDot && b then ". "^s
	else "("^ s ^")"
    in
    match e with
	Var i -> (if i >= 0 
		  then let (x, t) = cxt_lookup (gL, i) in x
		  else match (env_lookup (thL, i)) with
		      Def (x, tau, t, strict) -> x
		    | Decl (x, tau) -> x
		    | Cell c -> 
			match !c with
(*
			    Name (x, g', kappa) -> x
			  | Value (tau, g', kappa) -> tos (n, g', gR, tau))
*)
			    ChName (x, gL', gR', tau) -> x
			  | ChValue (e', gL', gR', tau) -> tos (b, n, gL', gR', e'))

      | C(TmApp, [[],e;[],e']) -> if n >= 10 then par ()
	else (tos (false, 9, gL, gR, e))^" "^(tos (b, 10, gL, gR, e'))

      | C(TyApp, [[],e;[],t]) -> 
	(tos (false, 11, gL, gR, e))^"["^(Clos.toString' (0, ((thR, gR), t)))^"]"

(* old version: print tyApp like tmApp
      | C(TyApp, [[],e;[],t]) -> if n >= 10 then par ()
	else (tos (9, gL, gR, e))^" "^(Clos.toString' (10, ((thR, gR), t)))
*)

      | C(TmAbs, [[],sigma; [x],e]) -> if n > 0 && (not b) then par ()
	else let x' = nextInst ((thL, gL), x)
	in Keyword.lam ^x' ^ ":" ^
	   (Clos.toString' (1, ((thR, gR), sigma))) ^
	   (abs (cxt_pushLit' (x', gL), gR, e))

      | C(TyAbs, [[],kappa; [a],e]) -> if n > 0 && (not b) then par ()
	else let a' = nextInst ((thR, gR), a)
	in Keyword.tyAbs ^a' ^ ":" ^
	   (Clos.toString' (1, (Clos.close (kappa)))) ^
	   (abs (gL, cxt_pushLit' (a', gR), e))

      | _ -> "MALFORMED EXPRESSION"

  and abs (gL, gR, e) = 
    match e with

      | C(TmAbs, [[],sigma; [x],e]) -> 
	let x' = nextInst ((thL, gL), x)
	in " " ^ Keyword.lam ^ x' ^ ":" ^
	   (Clos.toString' (1, ((thR, gR), sigma))) ^
	   (abs (cxt_pushLit' (x', gL), gR, e))

      | C(TyAbs, [[],kappa; [a],e]) -> 
	let a' = nextInst ((thR, gR), a)
	in " " ^ Keyword.tyAbs ^a' ^ ":" ^ 
	   (Clos.toString' (1, (Clos.close (kappa)))) ^
	   (abs (gL, cxt_pushLit' (a', gR), e))

      | _ -> ". " ^ tos (true, 0, gL, gR, e)
		  
  in
    tos (true, 0, gL0, gR0, e0)




(** Type inference for Church-style terms *)

exception IllFormed
exception NotAFunctionType
exception NotAUniversalType


let rec infer (((thL, gL0), t0), (thR, gR0)) =

  let rec inf (gL, t', gR) = 

    let return (tau) = 
      log ("exit_infer", fun () ->
	   "infer "^toString ((thL, gL), (thR, gR), t')^"{"^Exp.toString(t')^"}\n"^ 
	   "    : "^Clos.toString ((thR, gR), tau)^"\n");
      tau 
    in

    match t' with
	
	Var i -> 
	  (if i >= 0 then
	     let (x, tau) = cxt_lookupTmVar (gL, i, gR) in tau
	   else match (env_lookup (thL, i)) with
	       Def (x, tau, t, strict) -> return (tau)
	     | Decl (x, tau) -> return (tau)
	     | Cell c -> match !c with
(*
		   Name _ -> raise IllFormed
		 | Value (r, g', kappa) -> 
		     inf (g', r, gR)) (* bug *)
*)
		   ChValue (r, gL', gR', tau) -> 
		     inf (gL, churchPromote (r, (thL, gL'), gL, (thR, gR'), gR), gR)
		 | _ -> raise IllFormed)


      | C(TmAbs, [[],tau; [x],t]) ->
	  let gL' = cxt_pushTmVar (x, (gR, tau), (thL, gL))
	  in return (mkArr (tau, inf (gL', t, gR)))

      (* case: application  r s *)
      | C(TmApp, [[],r; [],s]) ->
	  (* head normal form of type of r should be sigma->tau *)
	  (match Norm.hnf ((thR, gR), inf (gL, r, gR)) with
	       C(Arr, [[],sigma; [],tau]) ->
		 (* ensure sigma == type of s and return tau *)
(* final code *)
		 check (((thL, gL), s), ((thR, gR), sigma), mkType);
(* old code for debugging 
		 let sigma' = inf (gL, s, gR)
		 in 
		   print_string ("r     = "^toString ((thL, gL), (thR, gR), r)^"["^Exp.toString (r)^"] : "^Clos.toString ((thR, gR), inf (gL, r, gR))^"\n");
		   print_string ("s     = "^toString ((thL, gL), (thR, gR), s)^"["^Exp.toString (s)^"] : "^Clos.toString ((thR, gR), inf (gL, s, gR))^"\n");
		   print_string ("sigma : "^Clos.toString ((thR, gR), sigma)^"\n");
		   print_string ("sigma': "^Clos.toString ((thR, gR), sigma')^"\n");
		   eq ((thR, gR), sigma, sigma', mkType); 
*)
		 return (tau)
	     | _ -> raise NotAFunctionType)	 

      | C(TyAbs, [[],kappa; [a],t]) ->
	  let gR' = cxt_pushTyVar (a, kappa, (thR, gR))
	  in return (mkForall (a, kappa, inf (gL, t, gR')))

      | C(TyApp, [[],r; [],sigma]) ->
	  (* head normal form of type of r should be forall a:kappa.tau *)
	  (match Norm.hnf ((thR, gR), inf (gL, r, gR)) with
	       C(Forall, [[],kappa; [a],tau]) ->

(* For now, we believe that sigma has the right kind

		 (* ensure that sigma has kind kappa *)
		 check (((thR, gR), sigma), Clos.close (kappa), mkKind);
*)
		 return (Pie.subst1 (tau, sigma))
	     | _ -> raise NotAUniversalType)

(* infer only works for terms and types, not for kinds
      | C(Type, []) -> mkKind
*)

      | C(Arr, [[],r; [],s]) -> 
(* Believe...
	  checkType ((thR, gR), r); 
	  checkType ((thR, gR), s); 
*)
	  mkType

      | C(Forall, [[],kappa; [a],tau]) ->
(* Believe...
	  let gR' = cxt_pushTyVar (a, kappa, (thR, gR))
	  in checkType ((thR, gR'), tau);
*)
	  mkType

      | _ -> raise IllFormed

  in (* begin infer *)
    inf (gL0, t0, gR0)

and check (((thL, gL), tL), ((thR, gR), tR), sort) =
  log ("enter_check", fun () -> 
	 "check "^toString ((thL, gL), (thR, gR), tL)^" : "^Clos.toString ((thR, gR), tR)^" : "^Exp.toString (sort)^"\n");

  eq ((thR, gR), tR, infer (((thL, gL), tL), (thR, gR)), sort);

  log ("exit_check", fun () -> 
	 "okchk "^toString ((thL, gL), (thR, gR), tL)^" : "^Clos.toString ((thR, gR), tR)^" : "^Exp.toString (sort)^"\n")


and checkType (c) = 
  check (c, Clos.close (mkType), mkKind)




(** Eta-reduction normal forms for Church terms *)


(** Auxiliary data structure of variable occurrence lists *)

type occList =
    Nil                        (** empty occurrence list *)
  | Cons of bool * occList     (** occ.list extended by info for Var 0 *)
  | Join of occList * occList  (** delayed pointwise lor of 2 occ.lists *)

(** [mkOccList : int -> occList] 
  generates a occurence list constant with i-1 times false and a true *)
let rec mkOccList (i) = 
  if i = 0 then Cons (true, Nil)
  else if i < 0 then raise (Error "cannot generate occurence list for negative deBruijn variable")
  else Cons (false, mkOccList (i-1))

(** [occHnf : occList -> occList{value} ]
  [occHnf (o) = o'] where [o'] is a value, i.e., a [Nil] or a [Cons] *)
let rec occHnf (o) =
  match o with
      Join (p, q) -> hJoin (occHnf p, occHnf q)
    | _ -> o

(** [hJoin : occList * occList -> occList]
  expects two values [(p, q)] and returns a value [o] *)
and hJoin (p, q) =
  match (p, q) with
      (Cons (b, p), Cons (c, q)) -> Cons (b || c, Join (p, q))
    | (Cons (b, p), Nil) -> Cons (b, Join (p, Nil))  (* preserve structure *)
    | (Nil, Cons (c, q)) -> Cons (c, Join (Nil, q))
    | (Nil, Nil) -> Cons (false, Join (Nil, Nil))


(** [occSplit : occList -> bool * occList] *)
let occSplit (o) =
  match (occHnf (o)) with
      Nil -> (false, Nil)
    | Cons (b, p) -> (b, p)

let rec join (p, q) =
  match (p, q) with
      (Cons (b, p), Cons (c, q)) -> Cons (b || c, join (p, q))
    | (_, Nil) -> p
    | (Nil, q) -> q

let rec occEval (o) =
  match o with
      Nil -> Nil
    | Cons (b, p) -> Cons (b, occEval p)
    | Join (p, q) -> join (occEval p, occEval q)


(** [delHead : occList -> occList] *)
let delHead (o) = 
  match occEval (o) with
      Nil -> Nil
    | Cons (b, p) -> p

let rec occListToString (o) =
  match o with
      Nil -> ""
    | Cons (true,  o') -> "1"^occListToString (o')
    | Cons (false, o') -> "0"^occListToString (o')
    | Join (p, q) -> "("^occListToString(p)^"|"^occListToString(q)^")"


(** curryEta : Clos.clos -> (occList * exp -> exp) -> exp *)

let curryEta ((th, g0), t0) k =

  let rec eta (g, (t:exp)) k =

    let atom () = k (Nil, t) in

    let appArr (r, s) f =
      eta (g, r)
	(fun (p, r') ->
	   eta (g, s) 
	   (fun (q, s') -> 
	      k (Join (p, q), f (r', s'))))
    in

    match t with

	Var i -> 
	  (if i >= 0 then k (mkOccList (i), t)
	   else match env_lookup (th, i) with
	     Def _ -> atom ()
	  | Decl _ -> atom ()
	  | Cell c -> match !c with
		Name _ -> atom ()
	      | Value (r, gX, tau) -> 
		  eta (g, promote (gX, r, g)) k 
	  )

      | C(Lam, [[x],r]) ->
	  let g' = cxt_push' (x, (th, g)) in
	    eta (g', r)
	      (fun (o, t') ->
		 let noRed (o') = 
		   k (o', mkLam (x, t'))
		 in match t' with
		     C(App, [[],r'; [],Var 0]) ->
		       let Join (p, _) = o in
		       let (b, p') = occSplit (p) in
			 if b then (* Var 0 occurs, no red. poss *)
			   noRed (p')
			 else (* Var 0 does not occur: strengthen *)
			   k (p', promote (g', r', g))
		   | _ -> noRed (delHead (o)))

      | C(Forall, [[],kappa; [a],r]) ->
	  let g' = cxt_pushTyVar (a, kappa, (th, g)) in
	    eta (g', r)
	      (fun (o, t') -> k (delHead (o), mkForall (a, kappa, t')))
	  
      | C(App, [[],r; [],s]) -> appArr (r, s) mkApp
      | C(Arr, [[],r; [],s]) -> appArr (r, s) mkArr

      | _ -> raise (Error "curryEta: illformed Curry-term")

  in (* begin curryEta *)
    eta (g0, t0) k

(** [etanf : Church.clos -> Exp.exp] *)

let etanf ((thL, gL0), (thR, gR0), t) =

  (** [eta: int -> cxt * cxt * exp -> (occList * occList * exp -> exp) -> exp] *)
  let rec eta ind (gL, gR, t) k =

    log ("enter_eta", fun () -> 
	   "eta "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t)));
    log ("enter_eta_deBruijn", fun () -> " ["^Exp.toString (t)^"]");
    log ("enter_eta", fun () -> "\n");

    let atom () = k (Nil, Nil, t) in
    match t with

	Var i -> 
	  (if i >= 0 then k (mkOccList i, Nil, t)
	   else match env_lookup (thL, i) with
	     Def _ -> atom ()
	  | Decl _ -> atom ()
	  | Cell c -> match !c with
(*
		Name _ -> atom ()
	      | Value (r, gX, tau) -> 
		  eta ind (gL, gR, churchPromote (r, (thL, gX), gL, (thR, gR), gR)) k 
		  (* bug here *)
*)
		ChName _ -> atom ()
	      | ChValue (r, gLX, gRX, tau) -> 
		  eta ind (gL, gR, churchPromote (r, (thL, gLX), gL, (thR, gRX), gR)) k 
	  )

      | C(TmAbs, [[],tau; [x],r]) ->
	  let gL' = cxt_pushTmVar (x, (gR, tau), (thL, gL)) in
	    eta (ind+indent) (gL', gR, r)
	      (fun (oL, oR, t') ->
		 let noRed (oL') = 
		   curryEta ((thR, gR), tau)
		     (fun (pR, tau') -> 
			k (oL', Join (oR, pR), mkTmAbs (x, tau', t')))
		 in match t' with
		     C(TmApp, [[],r'; [],Var 0]) ->
		       let t'a = mkTmAbs (x, tau, t') in 
			 log ("debug_eta", fun () -> 
				"try "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t'a))^"\n"^ 
				"ocL "^(String.make ind ' ')^occListToString (oL)^"\n"^ 
				"ocR "^(String.make ind ' ')^occListToString (oR)^"\n");
		       let Join (p, _) = oL in
		       let (b, p') = occSplit (p) in
			 log ("debug_eta", fun () -> 
				"lL= "^(String.make ind ' ')^occListToString (Cons (b, p'))^"\n");

			 if b then (* Var 0 occurs, no red. poss *)
			   noRed (p')
			 else (* Var 0 does not occur: strengthen *)
			   let Join (pR, _) = oR in
			   let t'b = churchPromote (r', (thL, gL'), gL, 
						        (thR, gR), gR) in
			     log ("red_eta", fun () ->
				    "red "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t'a))^"\n"^ 
				    "--> "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t'b))^"\n");
			   k (p', pR, t'b)
		   | _ -> noRed (delHead (oL)))

      | C(TmApp, [[],r; [],s]) ->
	  eta ind (gL, gR, r)
	  (fun (pL, pR, r') ->
	     eta ind (gL, gR, s) 
	     (fun (qL, qR, s') ->
		let t' = mkTmApp (r', s') in
		  log ("debug_eta", fun () ->
			 "tma "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t'))^"\n");		
		  k (Join (pL, qL), Join (pR, qR), t')))

      | C(TyAbs, [[],kappa; [a],r]) ->
	  let gR' = cxt_pushTyVar (a, kappa, (thR, gR)) in
	    eta (ind+indent) (gL, gR', r)
	      (fun (oL, oR, t') ->
		 let t'a = mkTyAbs (a, kappa, t') in
		 match t' with
		     C(TyApp, [[],r'; [],Var 0]) ->
		  log ("debug_eta", fun () -> 
			 "try "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t'a))^"\n"^ 
			 "ocL "^(String.make ind ' ')^occListToString (oL)^"\n"^ 
			 "ocR "^(String.make ind ' ')^occListToString (oR)^"\n");

		       let Join (p, _) = oR in
		       let (b, p') = occSplit (p) in
		  log ("debug_eta", fun () -> 
			 "lR= "^(String.make ind ' ')^occListToString (Cons (b, p'))^"\n");
			 if b then (* TyVar 0 occurs, no red. poss *)
			   k (oL, p', t'a)
			 else (* Var 0 does not occur: strengthen *)
			   let t'b = churchPromote (r', (thL, gL), gL, 
						        (thR, gR'), gR) in
			     log ("red_eta", fun () ->
				    "red "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t'a))^"\n"^ 
				    "--> "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t'b))^"\n");
			     k (oL, p', t'b)
		   | _ -> k (oL, delHead (oR), t'a))
		 
      | C(TyApp, [[],r; [],tau]) ->
	  eta ind (gL, gR, r)
	  (fun (oL, pR, r') ->
	     curryEta ((thR, gR), tau) 
	     (fun (qR, tau') -> 
		let t' = mkTyApp (r', tau') in
		  log ("debug_eta", fun () -> 
			 "tya "^(String.make ind ' ')^(toString ((thL, gL), (thR, gR), t'))^"\n");
		k (oL, Join (pR, qR), t')))

      | _ -> raise (Error "churchEta: illformed Church term")


  in (* begin etanf *)
    eta 0 (gL0, gR0, t) (fun (o, p, t') -> t')
