(* sig EXP *)

(** Abstract syntax constructors mk... as more readable interface to [pieTm]. *) 

(* Expressions *)

type expC = Lam | App | Arr | Forall | Type | Kind (* Curry  Style Terms *)
	  | TmAbs | TmApp | TyAbs | TyApp          (* Church Style Terms *)
type exp  = expC Pie.pieTm
	      
val wfKind : exp -> bool
  
val mkApp : exp * exp -> exp
val mkLam : string * exp -> exp

val mkArr : exp * exp -> exp
val mkForall : string * exp * exp -> exp

val mkType : exp
val mkKind : exp

(* Church-style term constructors *)

(** [mkTmAbs (x, tau, e)] *)
val mkTmAbs : string * exp * exp -> exp

(** [mkTyAbs (a, kappa, e)] *)
val mkTyAbs : string * exp * exp -> exp

(** [mkTmAbb (e1, e2)] *)
val mkTmApp : exp * exp -> exp

(** [mkTyApp (e, tau)] *)
val mkTyApp : exp * exp -> exp

(** raw expression printing *)
val toString : exp -> string

(** Lifting for Church-style terms.

Because in Church-style terms, term and type variables are distinguished, the [Pie.lift] function cannot be used for Church-terms.  [Pie.lift] count all binders as binders for all variables.  This would be ok if we used a mixed term/type context (actually the better choice), but since we have separate contexts we also need separate lifting. *)

val tmVarLift  : exp * int * int -> exp
val tmVarLift1 : exp -> exp

val tyVarLift  : exp * int * int -> exp
val tyVarLift1 : exp -> exp
