(** Normalization procedures. *)

open Pie
open Exp
open Surr
open Clos

let nf ((expand0, sideterms0), ((theta, gamma), t0)) =
  let modified = ref (false) in 
    
  (* nfDn 
     handle eliminations, unpack Evars *)

  let rec nfDn ((expand, side), g, t0) =
    
    let rec nfApp (s, t) =
      match s with
      	  C (Lam, [[x],r]) -> modified := true; 
	    nfUp ((expand, true), g, Pie.subst1 (r, t))
      	| _ -> if side then Exp.mkApp (s, nfUp ((false, side), g, t))
	  else Exp.mkApp (s, t)
	    
    and iter t =
      match t with
	  Var i -> if i<0 then
	    match env_lookup (theta, i) with
	      	Decl _ -> t
	      | Def (x, kappa, r, strict) -> if expand then iter r else t
	      | Cell c -> match !c with
		    Name _ -> t
		  | Value (r, g', kappa) -> 
		      modified := true; 
		      iter (promote (g', r, g))
      	  else t
      	| C (App, [[],r; [],s]) -> nfApp (nfDn ((expand, side), g, r), s)
      	| _ -> t
    in
      iter t0
      	
  (* nfUp
     handle constructors *)

  and nfUp ((expand, side), g, t0) =
    let t = nfDn ((expand, side), g, t0) in
      if side then
      	match t with
	    C (Lam, [[x],s]) -> mkLam (x, nfUp ((false, side), cxt_push' (x, (theta, g)), s))
      	  | C (Forall, [[],kappa; [a],tau]) -> mkForall (a, kappa, nfUp ((false, side), cxt_pushTyVar (a, kappa, (theta, g)), tau))
      	  | C (Arr, [[],r; [],s]) -> mkArr (nfUp ((false, side),  g, r), nfUp ((false, side), g, s))
      	  | C (Type, []) -> t
      	  | C (Kind, []) -> t
      	  | _ -> t
      else t
  in
(*
  let p1= "nf  "^Clos.toString ((theta, gamma), t0) in
  (* let _ = print_string p1; flush stdout in *)
*)
  let v = nfUp ((expand0, sideterms0), gamma, t0) in
(*
  let p2 = " ==> "^Clos.toString ((theta, gamma), v)^"\n" in
  (* let _ = print_string p2; flush stdout in *)
  let _ = if !modified then print_string (p1^p2) else () in
*)
    v


(* Weak normal form                                                 *)
(* Remove all open redices in a term, but do not expand definitions *)

(* wnfc : clos -> clos *)

(* expand = false, sideterms = true *)
let wnf  (((th, g), t) as c) =           nf ((false, true), c)
let wnfc (((th, g), t) as c) = ((th, g), nf ((false, true), c))


(* Head normal form                                                 *)
(* Expose outermost constructor, expanding definitions as needed    *)

(* Property: if c is in weak normal form, then also hnf c           *)

(* hnfc : clos -> clos *)

(* expand = true, sideterms = false *)
let hnf  (((th, g), t) as c) =           nf ((true, false), c)
let hnfc (((th, g), t) as c) = ((th, g), nf ((true, false), c))

