(** Functions for output, logging and debugging. *)

exception Error of string

(** Output *)

(** Width of identifier field in output. *)
val nameWidth : int

(** An empty string of length [nameWidth]. *)
val nameField : string

(** Format an identifier to given width. *)
val formatWidth : int -> string -> string

(** Format an identifier to width [nameWidth]. *)
val formatName : string -> string

(** Output Church terms (default). *)
val churchOn : unit -> unit

(** Do not output Church terms. *)
val churchOff: unit -> unit

(** Make extensive use of dot notation if flag set. *)
val extensiveDot : bool ref

val logLevel : int ref

(** Logging into logfile. *)

(** Indentation step. *)
val indent : int

(*
(* [startlog (filename, level, ids)]
    Opens file [filename] and sets current log level to [level]
    and enables messages with identities in [ids]. *)
*)

(* [startlog (filename, level, ids)]
    Opens file [filename] and starts logging into this file. *)
val startLog : string -> unit

(** [log (id, msg)] 
    Prints [msg] into current logfile if level of message [id]
    <= current debug level or [id] has been explicitely set for logging. *)
val log : string * (unit -> string) -> unit

(** Logging levels. 
  0 = mandatory
      main announcements, result of larger computations
  1 = nicely formatted, meaningful primary output 
      trace of type-checking : ass, chk
  2 = nicely formatted, meaningful secondary output 
      trace of type-checking : new, set, uni
      trace of unification   : ass, new, set
  3 = nicely formatted, meaningful ternary output  
      trace of chk/uni       : Church-terms
      double-checking        : eta-reductions
  4 = "entering function" messages for main functions
      e.g. CHK, leq
  5 = "entering function" messages for smaller functions
      e.g. 
  6 = everything
  
*)

val logLevel : int ref

(** Trace generation (includes only messages of level <= [traceLevel]) *)

val traceLevel : int ref

(** Clear old trace, start collecting log messages anew. *)
val beginTrace : unit -> unit

(** Print collected messages. *)
val printTrace : unit -> unit

(** Trace info to stderr if [debugOn()] is called *)
val debugOn : unit -> unit
