(* struct Top *)
(** Processing of declarations. *)

open Output
open Surr
open Global

let declTm (x, tau) = 
  let tauCl = closeTy (tau)
  in 
    log ("decl", fun () ->
	   "term "^formatName(x)^" : "^(Clos.toString tauCl)^";\n");
    log ("checking_wf_type", fun () ->
	   "\nChecking wellformedness of type "^(Clos.toString tauCl)^" .\n");
    ignore (Check.chkType (tauCl));
    if x = "_" then () else (
      log ("adding_term_decl", fun () ->
	     "\nAdding term constant "^x^" : "^(Clos.toString tauCl)^" .\n");
      addTmDecl (x, tau))

let declTy (x, kappa) = 
  let 
    kappaCl = closeTy (kappa)
  in    
    log ("decl", fun () -> 
	   "type "^formatName(x)^" : "^(Clos.toString kappaCl)^";\n");
    log ("adding_type_decl", fun () ->
	   "\nAdding type constant "^x^" : "^(Clos.toString kappaCl)^" .\n");
    ignore (Check.chkKind (kappa));
    if x = "_" then () else addTyDecl (x, kappa)

let defTm (x, tau, t) =    
  let tauCl = closeTy (tau) in 
  let tCl = closeTm (t) in 
    log ("decl", fun () -> 
	   "term "^formatName(x)^" : "^(Clos.toString tauCl)^"\n"^
	   "     " ^ nameField ^ " = "^(Clos.toString tCl)^";\n");
    log ("checking_wf_type", fun () ->
	   "\nChecking wellformedness of type "^(Clos.toString tauCl)^" .\n");
    ignore (Check.chkType (tauCl));
    beginTrace ();
    log ("adding_term_def", fun () ->
	   "\nAdding term definition "^x^" : "^(Clos.toString tauCl)^" = "^(Clos.toString tCl)^" .\n");
(*
    let tNf = Norm.wnfc (tCl) in 
      log ("normal_form", fun () ->
	     "Normal form: "^(Clos.toString tNf)^" .\n\n");
*)
      let (thR, (thL, e)) = Check.chk (tCl, tauCl) in
	log ("church", fun () ->
	       (formatWidth (5+nameWidth) "   % Church term")^" = "^
	       Church.toString ((thL, tmCxt()), (thR, tyCxt()), e)^"\n");
	if x = "_" then () else addTmDef (x, tau, t)

let defTy (x, kappa, tau) =
  let
    kappaCl = closeKi (kappa)
  in let
    tauCl = closeTy (tau)
  in
    log ("decl", fun () -> 
	   "type "^formatName(x)^" : "^(Clos.toString kappaCl)^"\n"^
	   "     " ^ nameField ^ " = "^(Clos.toString tauCl)^";\n");
    log ("adding_type_def", fun () ->
	   "\nAdding type definition "^x^" : "^(Clos.toString kappaCl)^" = "^(Clos.toString tauCl)^" .\n");
    ignore (Check.chkKind (kappa));
    ignore (Check.chkConstr (tauCl, kappa));
(*
  let 
    strict = Strict.strict (tau)
  in
    log ("strict", fun () ->
	   "Notational definition is "
	   ^(if strict then "" else "NOT ")^"strict.\n") ;
*)
    let strict = Occur.occDef (tau) in
      if strict then () 
      else log ("strict", fun () -> "Warning: Notational definitions has void abstractions.\n");
      if x = "_" then () else addTyDef (x, kappa, tau, strict)

let doDecl (d) =
  log ("decl", fun () -> "\n");
  beginTrace ();
  match d with

      Decl.Chk (t, tau) -> defTm ("_", tau, t)
(*
	let tcl = closeTm (t)
	and tauCl = closeTy (tau) in
	  print_string(Clos.toString(tcl)^" : "^Clos.toString(tauCl)^"\n");
	  log ("checking_wf_type", fun () ->
		 "\nChecking wellformedness of type "^(Clos.toString tauCl)^" .\n");
	  ignore (Check.chkType (tauCl));
	  beginTrace ();
	  log ("checking", fun () -> 
		 "\nChecking "^(Clos.toString tcl)^" : "^(Clos.toString tauCl)^" ...\n\n");
	  ignore (Check.chk (tcl, tauCl))   
*)
    | Decl.TmDecl (x, tau) -> declTm (x, tau)
    | Decl.TyDecl (x, kappa) -> declTy (x, kappa)
    | Decl.TmDef  (x, tau, t) -> defTm (x, tau, t)
    | Decl.TyDef  (x, kappa, tau) -> defTy (x, kappa, tau)
 
